@@ -16,11 +16,23 @@
 ** with lots of bits pasted from libpng.txt by Guy Eric Schalnat
 */
 
-#define VERSION "2.37.1 (3 July 1998)"
+#define VERSION "2.37.3 (13 July 1999)"
+
+/* GRR 19990713:  fixed redundant freeing of png_ptr and info_ptr in setjmp()
+ *  blocks and added "pm_close(ifp)" in each.  */
+
+/* GRR 19990317:  declared "clobberable" automatic variables in convertpng()
+ *  static to fix Solaris/gcc stack-corruption bug.  Also installed custom
+ *  error-handler to avoid jmp_buf size-related problems (i.e., jmp_buf
+ *  compiled with one size in libpng and another size here).  */
 
 #include <math.h>
 #include "pnm.h"
-#include "png.h"
+#include "png.h"	/* includes zlib.h and setjmp.h */
+
+typedef struct _jmpbuf_wrapper {
+  jmp_buf jmpbuf;
+} jmpbuf_wrapper;
 
 #ifndef TRUE
 #  define TRUE 1
@@ -35,10 +47,12 @@
 /* function prototypes */
 #ifdef __STDC__
 static png_uint_16 _get_png_val (png_byte **pp, int bit_depth);
-static void store_pixel (xel *pix, png_uint_16 r, png_uint_16 g, png_uint_16 b, png_uint_16 a);
+static void store_pixel (xel *pix, png_uint_16 r, png_uint_16 g, png_uint_16 b,
+                         png_uint_16 a);
 static int iscolor (png_color c);
 static void save_text (png_info *info_ptr, FILE *tfp);
 static void show_time (png_info *info_ptr);
+static void pngtopnm_error_handler (png_structp png_ptr, png_const_charp msg);
 static void convertpng (FILE *ifp, FILE *tfp);
 int main (int argc, char *argv[]);
 #endif
@@ -57,6 +71,7 @@ static float totalgamma = -1.0;
 static int text = FALSE;
 static char *text_file;
 static int mtime = FALSE;
+static jmpbuf_wrapper pngtopnm_jmpbuf_struct;
 
 #define get_png_val(p) _get_png_val (&(p), info_ptr->bit_depth)
 
@@ -180,6 +195,39 @@ png_info *info_ptr;
   }
 }
 
+#ifdef __STDC__
+static void pngtopnm_error_handler (png_structp png_ptr, png_const_charp msg)
+#else
+static void pngtopnm_error_handler (png_ptr, msg)
+png_structp png_ptr;
+png_const_charp msg;
+#endif
+{
+  jmpbuf_wrapper  *jmpbuf_ptr;
+
+  /* this function, aside from the extra step of retrieving the "error
+   * pointer" (below) and the fact that it exists within the application
+   * rather than within libpng, is essentially identical to libpng's
+   * default error handler.  The second point is critical:  since both
+   * setjmp() and longjmp() are called from the same code, they are
+   * guaranteed to have compatible notions of how big a jmp_buf is,
+   * regardless of whether _BSD_SOURCE or anything else has (or has not)
+   * been defined. */
+
+  fprintf(stderr, "pnmtopng:  fatal libpng error: %s\n", msg);
+  fflush(stderr);
+
+  jmpbuf_ptr = png_get_error_ptr(png_ptr);
+  if (jmpbuf_ptr == NULL) {         /* we are completely hosed now */
+    fprintf(stderr,
+      "pnmtopng:  EXTREMELY fatal error: jmpbuf unrecoverable; terminating.\n");
+    fflush(stderr);
+    exit(99);
+  }
+
+  longjmp(jmpbuf_ptr->jmpbuf, 1);
+}
+
 #define SIG_CHECK_SIZE 4
 
 #ifdef __STDC__
@@ -202,31 +250,45 @@ FILE *tfp;
   png_uint_16 c, c2, c3, a;
   int pnm_type;
   int i;
-  char *type_string = "";
-  char *alpha_string = "";
   int trans_mix;
   pixel backcolor;
   char gamma_string[80];
 
-  if (fread (sig_buf, 1, SIG_CHECK_SIZE, ifp) != SIG_CHECK_SIZE)
-    pm_error ("input file empty or too short");
-  if (png_sig_cmp (sig_buf, (png_size_t) 0, (png_size_t) SIG_CHECK_SIZE) != 0)
-    pm_error ("input file not a PNG file");
+  /* these variables are declared static because gcc wasn't kidding
+   * about "variable XXX might be clobbered by `longjmp' or `vfork'"
+   * (stack corruption observed on Solaris 2.6 with gcc 2.8.1, even
+   * in the absence of any other error condition) */
+  static char *type_string;
+  static char *alpha_string;
 
-  png_ptr = png_create_read_struct (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
+  type_string = alpha_string = "";
+
+  if (fread (sig_buf, 1, SIG_CHECK_SIZE, ifp) != SIG_CHECK_SIZE) {
+    pm_close (ifp);
+    pm_error ("input file empty or too short");
+  }
+  if (png_sig_cmp (sig_buf, (png_size_t) 0, (png_size_t) SIG_CHECK_SIZE) != 0) {
+    pm_close (ifp);
+    pm_error ("input file not a PNG file");
+  }
+
+  png_ptr = png_create_read_struct (PNG_LIBPNG_VER_STRING,
+    &pngtopnm_jmpbuf_struct, pngtopnm_error_handler, NULL);
   if (png_ptr == NULL) {
+    pm_close (ifp);
     pm_error ("cannot allocate LIBPNG structure");
   }
+
   info_ptr = png_create_info_struct (png_ptr);
   if (info_ptr == NULL) {
     png_destroy_read_struct (&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
+    pm_close (ifp);
     pm_error ("cannot allocate LIBPNG structures");
   }
 
-  if (setjmp (png_ptr->jmpbuf)) {
-    png_destroy_read_struct (&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
-    free (png_ptr);
-    free (info_ptr);
+  if (setjmp (pngtopnm_jmpbuf_struct.jmpbuf)) {
+    png_destroy_read_struct (&png_ptr, &info_ptr, (png_infopp)NULL);
+    pm_close (ifp);
     pm_error ("setjmp returns error condition");
   }
 
@@ -281,10 +343,9 @@ FILE *tfp;
 
   png_image = (png_byte **)malloc (info_ptr->height * sizeof (png_byte*));
   if (png_image == NULL) {
-    png_read_destroy (png_ptr, info_ptr, (png_info *)0);
-    free (png_ptr);
-    free (info_ptr);
-    pm_error ("couldn't alloc space for image");
+    png_destroy_read_struct (&png_ptr, &info_ptr, (png_infopp)NULL);
+    pm_close (ifp);
+    pm_error ("couldn't allocate space for image");
   }
 
   if (info_ptr->bit_depth == 16)
@@ -307,10 +368,9 @@ FILE *tfp;
       for (x = 0 ; x < y ; x++)
         free (png_image[x]);
       free (png_image);
-      png_read_destroy (png_ptr, info_ptr, (png_info *)0);
-      free (png_ptr);
-      free (info_ptr);
-      pm_error ("couldn't alloc space for image");
+      png_destroy_read_struct (&png_ptr, &info_ptr, (png_infopp)NULL);
+      pm_close (ifp);
+      pm_error ("couldn't allocate space for image");
     }
   }
 
@@ -499,10 +559,9 @@ FILE *tfp;
     for (y = 0 ; y < info_ptr->height ; y++)
       free (png_image[y]);
     free (png_image);
-    png_read_destroy (png_ptr, info_ptr, (png_info *)0);
-    free (png_ptr);
-    free (info_ptr);
-    pm_error ("couldn't alloc space for image");
+    png_destroy_read_struct (&png_ptr, &info_ptr, (png_infopp)NULL);
+    pm_close (ifp);
+    pm_error ("couldn't allocate space for image");
   }
 
   if (alpha == alpha_only) {
@@ -614,9 +673,8 @@ FILE *tfp;
           for (i = 0 ; i < info_ptr->height ; i++)
             free (png_image[i]);
           free (png_image);
-          png_read_destroy (png_ptr, info_ptr, (png_info *)0);
-          free (png_ptr);
-          free (info_ptr);
+          png_destroy_read_struct (&png_ptr, &info_ptr, (png_infopp)NULL);
+          pm_close (ifp);
           pm_error ("unknown PNG color type");
       }
       pnm_pixel++;
@@ -629,10 +687,7 @@ FILE *tfp;
   for (y = 0 ; y < info_ptr->height ; y++)
     free (png_image[y]);
   free (png_image);
-  png_read_destroy (png_ptr, info_ptr, (png_info *)0);
-
-  free (png_ptr);
-  free (info_ptr);
+  png_destroy_read_struct (&png_ptr, &info_ptr, (png_infopp)NULL);
 }
 
 #ifdef __STDC__
