@@ -17,9 +17,18 @@
 ** implied warranty.
 */
 
-#define VERSION "2.37.1 (3 July 1998)"
+#define VERSION "2.37.3 (13 July 1999)"
 
-/* GRR 980621:  moved some if-tests out of full-image loops; added fix for
+/* GRR 19990713:  fixed redundant freeing of png_ptr and info_ptr in setjmp()
+ *  blocks and added png_destroy_write_struct() and pm_close(ifp) in each
+ *  pm_error() block.  */
+
+/* GRR 19990308:  declared "clobberable" automatic variables in convertpnm()
+ *  static to fix Solaris/gcc stack-corruption bug.  Also installed custom
+ *  error-handler to avoid jmp_buf size-related problems (i.e., jmp_buf
+ *  compiled with one size in libpng and another size here).  */
+
+/* GRR 19980621:  moved some if-tests out of full-image loops; added fix for
  *  following problem discovered by Magnus Holmgren and debugged by Glenn:
  *
  *    The pnm has three colors:  0 0 0, 204 204 204, and 255 255 255.
@@ -36,10 +45,14 @@
 #define GRR_GRAY_PALETTE_FIX
 
 #include "pnm.h"
-#include "png.h"
+#include "png.h"	/* includes zlib.h and setjmp.h */
 
 #include "ppmcmap.h"
 
+typedef struct _jmpbuf_wrapper {
+  jmp_buf jmpbuf;
+} jmpbuf_wrapper;
+
 #ifndef TRUE
 #  define TRUE 1
 #endif
@@ -54,8 +67,10 @@
 
 /* function prototypes */
 #ifdef __STDC__
-static int closestcolor (pixel color, colorhist_vector chv, int colors, xelval maxval);
+static int closestcolor (pixel color, colorhist_vector chv, int colors,
+                         xelval maxval);
 static void read_text (png_info *info_ptr, FILE *tfp);
+static void pnmtopng_error_handler (png_structp png_ptr, png_const_charp msg);
 static void convertpnm (FILE *ifp, FILE *afp, FILE *tfp);
 int main (int argc, char *argv[]);
 #endif
@@ -92,6 +107,7 @@ static struct tm time_struct;
 static int filter = -1;
 static int compression = -1;
 static int force = FALSE;
+static jmpbuf_wrapper pnmtopng_jmpbuf_struct;
 
 #ifdef __STDC__
 static int closestcolor (pixel color, colorhist_vector chv, int colors, xelval maxval)
@@ -138,6 +154,7 @@ FILE *tfp;
   int c;
   char *cp;
 
+  /* GRR:  need to check for malloc failure here */
   info_ptr->text = (png_text *)malloc (MAXCOMMENTS * sizeof (png_text));
   j = 0;
   textpos = 0;
@@ -158,11 +175,11 @@ FILE *tfp;
         if (textline[0] == '"') {
           i++;
           while (textline[i] != '"' && textline[i] != '\n')
-            (*(cp++) = textline[i++]);
+            *(cp++) = textline[i++];
           i++;
         } else {
           while (textline[i] != ' ' && textline[i] != '\t' && textline[i] != '\n')
-            (*(cp++) = textline[i++]);
+            *(cp++) = textline[i++];
         }
         *(cp++) = '\0';
         cp = malloc (textpos);
@@ -172,7 +189,7 @@ FILE *tfp;
         strcpy (cp, &textline[i]);
         info_ptr->text[j].text_length = strlen (cp);
         j++;
-      }        else {
+      } else {
         j--;
         cp = malloc (info_ptr->text[j].text_length + textpos);
         strcpy (cp, info_ptr->text[j].text);
@@ -191,6 +208,39 @@ FILE *tfp;
   info_ptr->num_text = j;
 }
 
+#ifdef __STDC__
+static void pnmtopng_error_handler (png_structp png_ptr, png_const_charp msg)
+#else
+static void pnmtopng_error_handler (png_ptr, msg)
+png_structp png_ptr;
+png_const_charp msg;
+#endif
+{
+  jmpbuf_wrapper  *jmpbuf_ptr;
+
+  /* this function, aside from the extra step of retrieving the "error
+   * pointer" (below) and the fact that it exists within the application
+   * rather than within libpng, is essentially identical to libpng's
+   * default error handler.  The second point is critical:  since both
+   * setjmp() and longjmp() are called from the same code, they are
+   * guaranteed to have compatible notions of how big a jmp_buf is,
+   * regardless of whether _BSD_SOURCE or anything else has (or has not)
+   * been defined. */
+
+  fprintf(stderr, "pnmtopng:  fatal libpng error: %s\n", msg);
+  fflush(stderr);
+
+  jmpbuf_ptr = png_get_error_ptr(png_ptr);
+  if (jmpbuf_ptr == NULL) {         /* we are completely hosed now */
+    fprintf(stderr,
+      "pnmtopng:  EXTREMELY fatal error: jmpbuf unrecoverable; terminating.\n");
+    fflush(stderr);
+    exit(99);
+  }
+
+  longjmp(jmpbuf_ptr->jmpbuf, 1);
+}
+
 #ifdef __STDC__
 static void convertpnm (FILE *ifp, FILE *afp, FILE *tfp)
 #else
@@ -202,13 +252,11 @@ FILE *tfp;
 {
   xel **xels;	/* GRR:  a xel is always a pixel; pixel may be ulg or struct */
   xel p;	/*        (pnm.h)                  (ppm.h)  */
-  int rows, cols, format, pnm_type;
+  int rows, cols, format;
   xelval maxval;
-  xelval maxmaxval = 255;
   xelval scaleval;
   xelval value;
   pixel transcolor;
-  int sbitval;
   int mayscale;
   pixel backcolor;
 
@@ -222,42 +270,62 @@ FILE *tfp;
   png_byte *pp;
   int pass;
   int color;
-  gray **alpha_mask = NULL;	/* initialized to quiet compiler warnings */
   gray alpha_maxval;
   int alpha_rows;
   int alpha_cols;
-  int alpha_sbitval = 0;	/* initialized to quiet compiler warnings */
   int alpha_trans;
-  int num_alphas_of_color = 0;
   gray *alphas_of_color[MAXCOLORS];
   int alphas_of_color_cnt[MAXCOLORS];
   int alphas_first_index[MAXCOLORS+1];
   int mapping[MAXCOLORS];
-  int palette_size = 0;		/* initialized to quiet compiler warnings */
-  colorhist_vector chv;
-  colorhash_table cht = NULL;	/* initialized to quiet compiler warnings */
-  int depth=0, colors;		/* initialized to quiet compiler warnings */
+  int colors;
   int fulldepth;
   int x, y;
   int i, j;
 
-  png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
+  /* these variables are declared static because gcc wasn't kidding
+   * about "variable XXX might be clobbered by `longjmp' or `vfork'"
+   * (stack corruption observed on Solaris 2.6 with gcc 2.8.1, even
+   * in the absence of any other error condition) */
+  static int pnm_type;
+  static xelval maxmaxval;
+  static int sbitval;
+  static gray **alpha_mask;
+  static int alpha_sbitval;
+  static int num_alphas_of_color;
+  static int palette_size;
+  static colorhist_vector chv;
+  static colorhash_table cht;
+  static int depth;
+
+
+  /* these guys are initialized to quiet compiler warnings: */
+  maxmaxval = 255;
+  num_alphas_of_color = 0;
+  alpha_mask = NULL;
+  alpha_sbitval = 0;
+  palette_size = 0;
+  cht = NULL;
+  depth = 0;
+
+  png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING,
+    &pnmtopng_jmpbuf_struct, pnmtopng_error_handler, NULL);
   if (png_ptr == NULL) {
+    pm_close (ifp);
     pm_error ("cannot allocate LIBPNG structure");
   }
+
   info_ptr = png_create_info_struct (png_ptr);
   if (info_ptr == NULL) {
-    fclose(ifp);
     png_destroy_write_struct (&png_ptr, (png_infopp)NULL);
+    pm_close (ifp);
     pm_error ("cannot allocate LIBPNG structures");
   }
 
-  if (setjmp (png_ptr->jmpbuf)) {
-    fclose(ifp);
-    png_destroy_write_struct (&png_ptr, (png_infopp)NULL);
-    free (png_ptr);
-    free (info_ptr);
-    pm_error ("setjmp returns error condition");
+  if (setjmp (pnmtopng_jmpbuf_struct.jmpbuf)) {
+    png_destroy_write_struct (&png_ptr, &info_ptr);
+    pm_close (ifp);
+    pm_error ("setjmp returns error condition (1)");
   }
 
   xels = pnm_readpnm (ifp, &cols, &rows, &maxval, &format);
@@ -283,6 +351,8 @@ FILE *tfp;
   if (alpha) {
     alpha_mask = pgm_readpgm (afp, &alpha_cols, &alpha_rows, &alpha_maxval);
     if (alpha_cols != cols || alpha_rows != rows) {
+      png_destroy_write_struct (&png_ptr, &info_ptr);
+      pm_close (ifp);
       pm_error ("dimensions for image and alpha mask do not agree");
     }
     /* check if the alpha mask can be represented by a single transparency
@@ -357,8 +427,11 @@ FILE *tfp;
 
   sbitval = 0;
   if (pnm_type != PBM_TYPE || alpha) {
-    if (maxval > 65535 && !downscale)
+    if (maxval > 65535 && !downscale) {
+      png_destroy_write_struct (&png_ptr, &info_ptr);
+      pm_close (ifp);
       pm_error ("can only handle files up to 16-bit (use -downscale to override");
+    }
 
     if (maxval < 65536) {
       sbitval = pm_maxvaltobits (maxval);
@@ -612,7 +685,11 @@ FILE *tfp;
         for (i = 0 ; i < colors + 1 ; i++) {
           if ((alphas_of_color[i] = (gray *)malloc (MAXCOLORS * sizeof (int)))
               == NULL)
+          {
+            png_destroy_write_struct (&png_ptr, &info_ptr);
+            pm_close (ifp);
             pm_error ("out of memory allocating alpha/palette entries");
+          }
           alphas_of_color_cnt[i] = 0;
         }
         num_alphas_of_color = colors + 1;
@@ -723,8 +800,11 @@ FILE *tfp;
       depth = 2;
     else if (maxval == 1)
       depth = 1;
-    else
+    else {
+      png_destroy_write_struct (&png_ptr, &info_ptr);
+      pm_close (ifp);
       pm_error (" (can't happen) undefined maxval");
+    }
 
     if (alpha) {
       if (pnm_type == PPM_TYPE)
@@ -748,11 +828,10 @@ FILE *tfp;
 
   /* now write the file */
 
-  if (setjmp (png_ptr->jmpbuf)) {
-    png_write_destroy (png_ptr);
-    free (png_ptr);
-    free (info_ptr);
-    pm_error ("setjmp returns error condition");
+  if (setjmp (pnmtopng_jmpbuf_struct.jmpbuf)) {
+    png_destroy_write_struct (&png_ptr, &info_ptr);
+    pm_close (ifp);
+    pm_error ("setjmp returns error condition (2)");
   }
 
 #ifdef OLDPNG
@@ -846,9 +925,12 @@ FILE *tfp;
         }
       }
       /* indices should have just crossed paths */
-      if (bot_idx != top_idx + 1)
+      if (bot_idx != top_idx + 1) {
+        png_destroy_write_struct (&png_ptr, &info_ptr);
+        pm_close (ifp);
         pm_error ("internal inconsistency: remapped bot_idx = %d, top_idx = %d",
           bot_idx, top_idx);
+      }
       for (i = 0 ; i < colors ; i++) {
         for (j = alphas_first_index[i] ; j < alphas_first_index[i+1] ; j++) {
           palette[mapping[j]].red   = PPM_GETR (chv[i].color);
@@ -898,8 +980,11 @@ FILE *tfp;
         info_ptr->trans_values.blue = PPM_GETB (transcolor);
       }
     } else {
-      if (transparent > 0)
+      if (transparent > 0) {
+        png_destroy_write_struct (&png_ptr, &info_ptr);
+        pm_close (ifp);
         pm_error (" (can't happen) transparency AND alpha");
+      }
     }
 
   /* bKGD chunk */
@@ -991,8 +1076,13 @@ FILE *tfp;
   /* let libpng take care of, e.g., bit-depth conversions */
   png_set_packing (png_ptr);
 
-  /* GRR:  need to check for malloc failure here */
-  line = malloc (cols*8); /* max: 3 color channels, one alpha channel, 16-bit */
+  /* max: 3 color channels, one alpha channel, 16-bit */
+  if ((line = (png_byte *) malloc (cols*8)) == NULL)
+  {
+    png_destroy_write_struct (&png_ptr, &info_ptr);
+    pm_close (ifp);
+    pm_error ("out of memory allocating PNG row buffer");
+  }
 
   for (pass = 0 ; pass < png_set_interlace_handling (png_ptr) ; pass++) {
     for (y = 0 ; y < rows ; y++) {
@@ -1007,8 +1097,11 @@ FILE *tfp;
         } else if (info_ptr->color_type == PNG_COLOR_TYPE_PALETTE) {
           color = ppm_lookupcolor (cht, &p);
           if (alpha) {
-            for (i = alphas_first_index[color] ; i < alphas_first_index[color+1] ; i++)
-              if (alpha_mask[y][x] == alphas_of_color[color][i - alphas_first_index[color]]) {
+            for (i = alphas_first_index[color] ;
+                 i < alphas_first_index[color+1] ; i++)
+              if (alpha_mask[y][x] ==
+                  alphas_of_color[color][i - alphas_first_index[color]])
+              {
                 color = mapping[i];
                 break;
               }
@@ -1026,6 +1119,8 @@ FILE *tfp;
             *pp++ = PPM_GETB (p) >> 8;
           *pp++ = PPM_GETB (p)&0xff;
         } else {
+          png_destroy_write_struct (&png_ptr, &info_ptr);
+          pm_close (ifp);
           pm_error (" (can't happen) undefined color_type");
         }
         if (info_ptr->color_type & PNG_COLOR_MASK_ALPHA) {
@@ -1192,7 +1287,8 @@ char *argv[];
         sscanf (argv[argn], "%d", &filter);
         if ((filter < 0) || (filter > 4))
         {
-          pm_message ("filter must be 0 (none), 1 (sub), 2 (up), 3 (avg) or 4 (paeth)");
+          pm_message
+            ("filter must be 0 (none), 1 (sub), 2 (up), 3 (avg) or 4 (paeth)");
           pm_usage (usage);
         }
       }
